@@ -23,12 +23,19 @@
 #include <iostream>
 #include <cstring>
 #include <cstdlib>
+#include <exception>
 
 /* Need to know which geos version we have to work out which headers to include */
 #include <geos/version.h>
 
+/* geos (3.0.0+) */
 #if (GEOS_VERSION_MAJOR==3)
-/* geos trunk (3.0.0+) */
+#if (GEOS_VERSION_MINOR>=1)
+/* Prepared geometries are new in 3.1.0 */
+#define HAS_PREPARED_GEOMETRIES
+#include <geos/geom/prep/PreparedGeometryFactory.h>
+#include <geos/geom/prep/PreparedPolygon.h>
+#endif
 #include <geos/geom/GeometryFactory.h>
 #include <geos/geom/CoordinateSequenceFactory.h>
 #include <geos/geom/Geometry.h>
@@ -90,9 +97,15 @@ char *get_wkt_simple(osmNode *nodes, int count, int polygon) {
         std::string wkt = wktw.write(geom.get());
         return strdup(wkt.c_str());
     }
+    catch (std::bad_alloc)
+    {
+        std::cerr << std::endl << "Exception caught processing way. You are likelly running out of memory." << std::endl;
+        std::cerr << "Try in slim mode, using -s parameter." << std::endl;
+        return NULL;
+    }
     catch (...)
     {
-        std::cerr << std::endl << "excepton caught processing way" << std::endl;
+        std::cerr << std::endl << "Exception caught processing way" << std::endl;
         return NULL;
     }
 }
@@ -149,9 +162,15 @@ size_t get_wkt_split(osmNode *nodes, int count, int polygon, double split_at) {
         }
 
     }
+    catch (std::bad_alloc)
+    {
+        std::cerr << std::endl << "Exception caught processing way. You are likelly running out of memory." << std::endl;
+        std::cerr << "Try in slim mode, using -s parameter." << std::endl;
+        wkt_size = 0;
+    }
     catch (...)
     {
-        std::cerr << std::endl << "excepton caught processing way" << std::endl;
+        std::cerr << std::endl << "Exception caught processing way" << std::endl;
         wkt_size = 0;
     }
     return wkt_size;
@@ -253,7 +272,7 @@ int parse_wkt(const char * wkt, struct osmNode *** xnodes, int ** xcount, int *
         }
         delete geometry;
     } catch (...) {
-        std::cerr << std::endl << "excepton caught parsing PostGIS data" << std::endl;
+        std::cerr << std::endl << "Exception caught parsing PostGIS data" << std::endl;
         return -1;
     }
     return 0;
@@ -283,6 +302,9 @@ size_t build_geometry(int osm_id, struct osmNode **xnodes, int *xcount, int make
     std::auto_ptr<std::vector<Geometry*> > lines(new std::vector<Geometry*>);
     GeometryFactory gf;
     geom_ptr geom;
+#ifdef HAS_PREPARED_GEOMETRIES
+    geos::geom::prep::PreparedGeometryFactory pgf;
+#endif
 
     try
     {
@@ -370,11 +392,19 @@ size_t build_geometry(int osm_id, struct osmNode **xnodes, int *xcount, int make
             {
                 if (polys[i].iscontained != 0) continue;
                 toplevelpolygons++;
+#ifdef HAS_PREPARED_GEOMETRIES
+                const geos::geom::prep::PreparedGeometry* preparedtoplevelpolygon = pgf.create(polys[i].polygon);
+#endif
 
                 for (unsigned j=i+1; j < totalpolys; ++j)
                 {
+#ifdef HAS_PREPARED_GEOMETRIES
+                    // Does preparedtoplevelpolygon contain the smaller polygon[j]?
+                    if (polys[j].containedbyid == 0 && preparedtoplevelpolygon->contains(polys[j].polygon))
+#else
                     // Does polygon[i] contain the smaller polygon[j]?
                     if (polys[j].containedbyid == 0 && polys[i].polygon->contains(polys[j].polygon))
+#endif
                     {
                         // are we in a [i] contains [k] contains [j] situation
                         // which would actually make j top level
@@ -386,9 +416,9 @@ size_t build_geometry(int osm_id, struct osmNode **xnodes, int *xcount, int make
                                 istoplevelafterall = 1;
                                 break;
                             }
+#if 0
                             else if (polys[k].polygon->intersects(polys[j].polygon) || polys[k].polygon->touches(polys[j].polygon))
 			    {
-#if 0
                                 // FIXME: This code does not work as intended
                                 // It should be setting the polys[k].ring in order to update this object
                                 // but the value of polys[k].polygon calculated is normally NULL
@@ -401,8 +431,9 @@ size_t build_geometry(int osm_id, struct osmNode **xnodes, int *xcount, int make
 				polys[j].iscontained = 2; // Drop
                                 istoplevelafterall = 2;
                                 break;
-#endif
+
                             }
+#endif
                         }
                         if (istoplevelafterall == 0)
                         {
@@ -411,6 +442,9 @@ size_t build_geometry(int osm_id, struct osmNode **xnodes, int *xcount, int make
                         }
                     }
                 }
+#ifdef HAS_PREPARED_GEOMETRIES
+		pgf.destroy(preparedtoplevelpolygon);
+#endif
             }
             // polys now is a list of ploygons tagged with which ones are inside each other
 
@@ -466,11 +500,16 @@ size_t build_geometry(int osm_id, struct osmNode **xnodes, int *xcount, int make
         }
         delete[](polys);
     }
+    catch (std::exception& e)
+      {
+	std::cerr << std::endl << "Standard exception processing way_id "<< osm_id << ": " << e.what()  << std::endl;
+	wkt_size = 0;
+      }
     catch (...)
-    {
-        std::cerr << std::endl << "excepton caught processing way id=" << osm_id << std::endl;
+      {
+        std::cerr << std::endl << "Exception caught processing way id=" << osm_id << std::endl;
         wkt_size = 0;
-    }
+      }
 
     return wkt_size;
 }
