@@ -29,4 +29,6 @@ void resetList(struct keyval *head);
 struct keyval *getMatches(struct keyval *head, const char *name);
 void updateItem(struct keyval *head, const char *name, const char *value);
 void cloneList( struct keyval *target, struct keyval *source );
+void keyval2hstore(char *hstring, struct keyval *tags);
+void keyval2hstore_manual(char *hstring, char *key, char *value);
 #endif
