@@ -6,6 +6,8 @@
  * emit the final geometry-enabled output formats
 */
 
+#include "config.h"
+
 #include <stdio.h>
 #include <unistd.h>
 #include <stdlib.h>
@@ -52,10 +54,10 @@ static struct s_table {
     int copyMode;
     char *columns;
 } tables [] = {
-    { name: "%s_point",   type: "POINT"     },
-    { name: "%s_line",    type: "LINESTRING"},
-    { name: "%s_polygon", type: "GEOMETRY"  }, // Actually POLGYON & MULTIPOLYGON but no way to limit to just these two
-    { name: "%s_roads",   type: "LINESTRING"}
+    { .name = "%s_point",   .type = "POINT"     },
+    { .name = "%s_line",    .type = "LINESTRING"},
+    { .name = "%s_polygon", .type = "GEOMETRY"  }, // Actually POLGYON & MULTIPOLYGON but no way to limit to just these two
+    { .name = "%s_roads",   .type = "LINESTRING"}
 };
 #define NUM_TABLES ((signed)(sizeof(tables) / sizeof(tables[0])))
 
@@ -63,14 +65,16 @@ static struct s_table {
 #define FLAG_LINEAR  2    /* For lines table */
 #define FLAG_NOCACHE 4    /* Optimisation: don't bother remembering this one */
 #define FLAG_DELETE  8    /* These tags should be simply deleted on sight */
+#define FLAG_PHSTORE 17   /* polygons without own column but listed in hstore this implies FLAG_POLYGON */
 static struct flagsname {
     char *name;
     int flag;
 } tagflags[] = {
-    { name: "polygon",    flag: FLAG_POLYGON },
-    { name: "linear",     flag: FLAG_LINEAR },
-    { name: "nocache",    flag: FLAG_NOCACHE },
-    { name: "delete",     flag: FLAG_DELETE }
+    { .name = "polygon",    .flag = FLAG_POLYGON },
+    { .name = "linear",     .flag = FLAG_LINEAR },
+    { .name = "nocache",    .flag = FLAG_NOCACHE },
+    { .name = "delete",     .flag = FLAG_DELETE },
+    { .name = "phstore",    .flag = FLAG_PHSTORE }
 };
 #define NUM_FLAGS ((signed)(sizeof(tagflags) / sizeof(tagflags[0])))
 
@@ -139,7 +143,7 @@ void read_style_file( const char *filename )
     lineno++;
     
     char osmtype[24];
-    char tag[24];
+    char tag[64];
     char datatype[24];
     char flags[128];
     int i;
@@ -149,7 +153,7 @@ void read_style_file( const char *filename )
     if( str )
       *str = '\0';
       
-    int fields = sscanf( buffer, "%23s %23s %23s %127s", osmtype, tag, datatype, flags );
+    int fields = sscanf( buffer, "%23s %63s %23s %127s", osmtype, tag, datatype, flags );
     if( fields <= 0 )  /* Blank line */
       continue;
     if( fields < 3 )
@@ -175,6 +179,13 @@ void read_style_file( const char *filename )
       if( i == NUM_FLAGS )
         fprintf( stderr, "Unknown flag '%s' line %d, ignored\n", str, lineno );
     }
+    if (temp.flags==FLAG_PHSTORE) {
+      if (0==(Options->enable_hstore)) {
+	fprintf( stderr, "Error reading style file line %d (fields=%d)\n", lineno, fields );
+	fprintf( stderr, "flag 'phstore' is invalid in non-hstore mode\n");
+	exit_nicely();
+      }
+    }
     temp.count = 0;
 //    printf("%s %s %d %d\n", temp.name, temp.type, temp.polygon, offset );
     
@@ -427,6 +438,137 @@ static void escape_type(char *sql, int len, const char *value, const char *type)
     }
 }
 
+static void write_hstore(enum table_id table, struct keyval *tags)
+{
+    static char *sql;
+    static size_t sqllen=0;
+    
+    // sql buffer
+    if (sqllen==0) {
+      sqllen=2048;
+      sql=malloc(sqllen);
+    }
+    
+    // a clone of the tags pointer
+    struct keyval *xtags = tags;
+    
+    // while this tags has a follow-up..
+    while (xtags->next->key != NULL)
+    {
+        /*
+          hstore ASCII representation looks like
+          "<key>"=>"<value>"
+          
+          we need at least strlen(key)+strlen(value)+6+'\0' bytes
+          in theory any single character could also be escaped
+          thus we need an additional factor of 2.
+          The maximum lenght of a single hstore element is thus
+          calcuated as follows:
+        */
+        size_t hlen=2 * (strlen(xtags->next->key) + strlen(xtags->next->value)) + 7;
+        
+        // if the sql buffer is too small
+        if (hlen > sqllen) {
+            sqllen = hlen;
+            sql = realloc(sql, sqllen);
+        }
+        
+        // pack the tag with its value into the hstore
+        keyval2hstore(sql, xtags->next);
+        copy_to_table(table, sql);
+        
+        // update the tag-pointer to point to the next tag
+        xtags = xtags->next;
+        
+        // if the tag has a follow up, add a comma to the end
+        if (xtags->next->key != NULL)
+            copy_to_table(table, ",");
+    }
+    
+    // finish the hstore column by placing a TAB into the data stream
+    copy_to_table(table, "\t");
+    
+    // the main hstore-column has now been written
+}
+
+// write an hstore column to the database
+static void write_hstore_columns(enum table_id table, struct keyval *tags)
+{
+    static char *sql;
+    static size_t sqllen=0;
+    
+    // sql buffer
+    if (sqllen==0) {
+      sqllen=2048;
+      sql=malloc(sqllen);
+    }
+    
+    // the index of the current hstore column
+    int i_hstore_column;
+    
+    // iterate over all configured hstore colums in the options
+    for(i_hstore_column = 0; i_hstore_column < Options->n_hstore_columns; i_hstore_column++)
+    {
+        // did this node have a tag that matched the current hstore column
+        int found = 0;
+        
+        // a clone of the tags pointer
+        struct keyval *xtags = tags;
+        
+        // while this tags has a follow-up..
+        while (xtags->next->key != NULL) {
+            
+            // check if the tags' key starts with the name of the hstore column
+            char *pos = strstr(xtags->next->key, Options->hstore_columns[i_hstore_column]);
+            
+            // and if it does..
+            if(pos == xtags->next->key)
+            {
+                // remember we found one
+                found=1;
+                
+                // generate the short key name
+                char *shortkey = xtags->next->key + strlen(Options->hstore_columns[i_hstore_column]);
+                
+                // calculate the size needed for this hstore entry
+                size_t hlen=2*(strlen(shortkey)+strlen(xtags->next->value))+7;
+                
+                // if the sql buffer is too small
+                if (hlen > sqllen) {
+                    // resize it
+                    sqllen=hlen;
+                    sql=realloc(sql,sqllen);
+                }
+                
+                // and pack the shortkey with its value into the hstore
+                keyval2hstore_manual(sql, shortkey, xtags->next->value);
+                copy_to_table(table, sql);
+                
+                // update the tag-pointer to point to the next tag
+                xtags=xtags->next;
+                
+                // if the tag has a follow up, add a comma to the end
+                if (xtags->next->key != NULL)
+                    copy_to_table(table, ",");
+            }
+            else
+            {
+                // update the tag-pointer to point to the next tag
+                xtags=xtags->next;
+            }
+        }
+        
+        // if no matching tag has been found, write a NULL
+        if(!found)
+            copy_to_table(table, "\\N");
+        
+        // finish the hstore column by placing a TAB into the data stream
+        copy_to_table(table, "\t");
+    }
+    
+    // all hstore-columns have now been written
+}
+
 
 /* example from: pg_dump -F p -t planet_osm gis
 COPY planet_osm (osm_id, name, place, landuse, leisure, "natural", man_made, waterway, highway, railway, amenity, tourism, learning, building, bridge, layer, way) FROM stdin;
@@ -443,9 +585,17 @@ Workaround - output SRID=4326;<WKB>
 
 static int pgsql_out_node(int id, struct keyval *tags, double node_lat, double node_lon)
 {
-    char sql[2048], *v;
+
+    static char *sql;
+    static size_t sqllen=0;
+    char *v;
     int i;
 
+    if (sqllen==0) {
+      sqllen=2048;
+      sql=malloc(sqllen);
+    }
+
     expire_tiles_from_bbox(node_lon, node_lat, node_lon, node_lat);
     sprintf(sql, "%d\t", id);
     copy_to_table(t_point, sql);
@@ -453,9 +603,11 @@ static int pgsql_out_node(int id, struct keyval *tags, double node_lat, double n
     for (i=0; i < exportListCount[OSMTYPE_NODE]; i++) {
         if( exportList[OSMTYPE_NODE][i].flags & FLAG_DELETE )
             continue;
+	if( (exportList[OSMTYPE_NODE][i].flags & FLAG_PHSTORE) == FLAG_PHSTORE)
+            continue;
         if ((v = getItem(tags, exportList[OSMTYPE_NODE][i].name)))
         {
-            escape_type(sql, sizeof(sql), v, exportList[OSMTYPE_NODE][i].type);
+            escape_type(sql, sqllen, v, exportList[OSMTYPE_NODE][i].type);
             exportList[OSMTYPE_NODE][i].count++;
         }
         else
@@ -464,7 +616,14 @@ static int pgsql_out_node(int id, struct keyval *tags, double node_lat, double n
         copy_to_table(t_point, sql);
         copy_to_table(t_point, "\t");
     }
-
+    
+    // hstore columns
+    write_hstore_columns(t_point, tags);
+    
+    // check if a regular hstore is requested
+    if (Options->enable_hstore)
+        write_hstore(t_point, tags);
+    
     sprintf(sql, "SRID=%d;POINT(%.15g %.15g)", SRID, node_lon, node_lat);
     copy_to_table(t_point, sql);
     copy_to_table(t_point, "\n");
@@ -476,20 +635,29 @@ static int pgsql_out_node(int id, struct keyval *tags, double node_lat, double n
 
 static void write_wkts(int id, struct keyval *tags, const char *wkt, enum table_id table)
 {
+  
+    static char *sql;
+    static size_t sqllen=0;
+    char *v;
     int j;
-    char sql[2048];
-    const char*v;
 
+    if (sqllen==0) {
+      sqllen=2048;
+      sql=malloc(sqllen);
+    }
+    
     sprintf(sql, "%d\t", id);
     copy_to_table(table, sql);
 
     for (j=0; j < exportListCount[OSMTYPE_WAY]; j++) {
             if( exportList[OSMTYPE_WAY][j].flags & FLAG_DELETE )
                 continue;
+	    if( (exportList[OSMTYPE_WAY][j].flags & FLAG_PHSTORE) == FLAG_PHSTORE)
+                continue;
             if ((v = getItem(tags, exportList[OSMTYPE_WAY][j].name)))
             {
                 exportList[OSMTYPE_WAY][j].count++;
-                escape_type(sql, sizeof(sql), v, exportList[OSMTYPE_WAY][j].type);
+                escape_type(sql, sqllen, v, exportList[OSMTYPE_WAY][j].type);
             }
             else
                 sprintf(sql, "\\N");
@@ -497,39 +665,20 @@ static void write_wkts(int id, struct keyval *tags, const char *wkt, enum table_
             copy_to_table(table, sql);
             copy_to_table(table, "\t");
     }
-
+    
+    // hstore columns
+    write_hstore_columns(table, tags);
+    
+    // check if a regular hstore is requested
+    if (Options->enable_hstore)
+        write_hstore(table, tags);
+    
     sprintf(sql, "SRID=%d;", SRID);
     copy_to_table(table, sql);
     copy_to_table(table, wkt);
     copy_to_table(table, "\n");
 }
 
-void add_parking_node(int id, struct keyval *tags, double node_lat, double node_lon)
-{
-// insert into planet_osm_point(osm_id,name,amenity,way) select osm_id,name,amenity,centroid(way) from planet_osm_polygon where amenity='parking';
-	const char *access  = getItem(tags, "access");
-	const char *amenity = getItem(tags, "amenity");
-	const char *name    = getItem(tags, "name");
-	struct keyval nodeTags;
-	
-	if (!amenity || strcmp(amenity, "parking"))
-		return;
-        
-	// Do not add a 'P' symbol if access is defined and something other than public.
-	if (access && strcmp(access, "public"))
-		return;
-
-	initList(&nodeTags);
-	addItem(&nodeTags, "amenity", amenity, 0);
-	if (name)
-		addItem(&nodeTags, "name",    name,    0);
-	if (access)
-		addItem(&nodeTags, "access",  access,    0);
-	
-	pgsql_out_node(id, &nodeTags, node_lat, node_lon);
-	resetList(&nodeTags);
-}
-
 static int tag_indicates_polygon(enum OsmType type, const char *key)
 {
     int i;
@@ -561,14 +710,19 @@ unsigned int pgsql_filter_tags(enum OsmType type, struct keyval *tags, int *poly
     /* We used to only go far enough to determine if it's a polygon or not, but now we go through and filter stuff we don't need */
     while( (item = popItem(tags)) != NULL )
     {
-        /* Discard natural=coastline tags (we render these from a shapefile instead) */
+        /* Allow named islands to appear as polygons */
         if (!strcmp("natural",item->key) && !strcmp("coastline",item->value))
-        {		
+        {               
+            add_area_tag = 1; 
+        }
+
+        /* Discard natural=coastline tags (we render these from a shapefile instead) */
+        if (!Options->keep_coastlines && !strcmp("natural",item->key) && !strcmp("coastline",item->value))
+        {               
             freeItem( item );
             item = NULL;
-            add_area_tag = 1; /* Allow named islands to appear as polygons */
             continue;
-        }    
+        }
 
         for (i=0; i < exportListCount[type]; i++)
         {
@@ -591,8 +745,13 @@ unsigned int pgsql_filter_tags(enum OsmType type, struct keyval *tags, int *poly
         }
         if( i == exportListCount[type] )
         {
-            freeItem( item );
-            item = NULL;
+	  if (Options->enable_hstore) {
+	    pushItem( &temp, item );
+	    filter=0;
+	  } else {
+	    freeItem( item );
+	  }
+	  item = NULL;
         }
     }
 
@@ -811,6 +970,19 @@ static int pgsql_out_relation(int id, struct keyval *rel_tags, struct osmNode **
             }
         }
     }
+    else if( strcmp( type, "boundary" ) == 0 )
+    {
+        // Boundaries will get converted into multiple geometries:
+        // - Linear features will end up in the line and roads tables (useful for admin boundaries)
+        // - Polygon features also go into the polygon table (useful for national_forests)
+        // The edges of the polygon also get treated as linear fetaures allowing these to be rendered seperately.
+        make_boundary = 1;
+    }
+    else if( strcmp( type, "multipolygon" ) == 0 && getItem(&tags, "boundary") )
+    {
+        // Treat type=multipolygon exactly like type=boundary if it has a boundary tag.
+        make_boundary = 1;
+    }
     else if( strcmp( type, "multipolygon" ) == 0 )
     {
         make_polygon = 1;
@@ -841,14 +1013,6 @@ static int pgsql_out_relation(int id, struct keyval *rel_tags, struct osmNode **
             p = p->next;
         }
     }
-    else if( strcmp( type, "boundary" ) == 0 )
-    {
-        // Boundaries will get converted into multiple geometries:
-        // - Linear features will end up in the line and roads tables (useful for admin boundaries)
-        // - Polygon features also go into the polygon table (useful for national_forests)
-        // The edges of the polygon also get treated as linear fetaures allowing these to be rendered seperately.
-        make_boundary = 1;
-    }
     else
     {
         /* Unknown type, just exit */
@@ -961,7 +1125,7 @@ static int pgsql_out_relation(int id, struct keyval *rel_tags, struct osmNode **
 
 static int pgsql_out_start(const struct output_options *options)
 {
-    char *sql, tmp[128];
+    char *sql, tmp[256];
     PGresult   *res;
     int i,j;
     unsigned int sql_len;
@@ -970,7 +1134,7 @@ static int pgsql_out_start(const struct output_options *options)
 
     read_style_file( options->style );
 
-    sql_len = 1024;
+    sql_len = 2048;
     sql = malloc(sql_len);
     assert(sql);
 
@@ -1027,6 +1191,8 @@ static int pgsql_out_start(const struct output_options *options)
             for (j=0; j < numTags; j++) {
                 if( exportTags[j].flags & FLAG_DELETE )
                     continue;
+		if( (exportTags[j].flags & FLAG_PHSTORE ) == FLAG_PHSTORE)
+		    continue;
                 sprintf(tmp, ",\"%s\" %s", exportTags[j].name, exportTags[j].type);
                 if (strlen(sql) + strlen(tmp) + 1 > sql_len) {
                     sql_len *= 2;
@@ -1035,14 +1201,26 @@ static int pgsql_out_start(const struct output_options *options)
                 }
                 strcat(sql, tmp);
             }
-            strcat(sql, " );\n");
+            int i_hstore_column;
+            for(i_hstore_column = 0; i_hstore_column < Options->n_hstore_columns; i_hstore_column++)
+            {
+                strcat(sql, ",\"");
+                strcat(sql, Options->hstore_columns[i_hstore_column]);
+                strcat(sql, "\" hstore ");
+            }
+            if (Options->enable_hstore) {
+                strcat(sql, ",tags hstore );\n");
+	    } else {
+	      strcat(sql, " );\n");
+	    }
+
             pgsql_exec(sql_conn, PGRES_COMMAND_OK, "%s", sql);
             pgsql_exec(sql_conn, PGRES_TUPLES_OK, "SELECT AddGeometryColumn('%s', 'way', %d, '%s', 2 );\n",
                         tables[i].name, SRID, tables[i].type );
             pgsql_exec(sql_conn, PGRES_COMMAND_OK, "ALTER TABLE %s ALTER COLUMN way SET NOT NULL;\n", tables[i].name);
             /* slim mode needs this to be able to apply diffs */
             if( Options->slim )
-                pgsql_exec(sql_conn, PGRES_COMMAND_OK, "CREATE INDEX %s_pkey ON %s USING BTREE (osm_id);\n", tables[i].name, tables[i].name);
+	      pgsql_exec(sql_conn, PGRES_COMMAND_OK, "CREATE INDEX %s_pkey ON %s USING BTREE (osm_id) TABLESPACE %s;\n", tables[i].name, tables[i].name, Options->tblsindex);
         } else {
             /* Add any new columns referenced in the default.style */
             PGresult *res;
@@ -1055,6 +1233,8 @@ static int pgsql_out_start(const struct output_options *options)
             for (j=0; j < numTags; j++) {
                 if( exportTags[j].flags & FLAG_DELETE )
                     continue;
+		if( (exportTags[j].flags & FLAG_PHSTORE) == FLAG_PHSTORE)
+		    continue;
                 sprintf(tmp, "\"%s\"", exportTags[j].name);
                 if (PQfnumber(res, tmp) < 0) {
 #if 0
@@ -1081,6 +1261,8 @@ static int pgsql_out_start(const struct output_options *options)
         for (j=0; j < numTags; j++) {
             if( exportTags[j].flags & FLAG_DELETE )
                 continue;
+	    if( (exportTags[j].flags & FLAG_PHSTORE ) == FLAG_PHSTORE)
+		    continue;
             sprintf(tmp, ",\"%s\"", exportTags[j].name);
 
             if (strlen(sql) + strlen(tmp) + 1 > sql_len) {
@@ -1090,7 +1272,18 @@ static int pgsql_out_start(const struct output_options *options)
             }
             strcat(sql, tmp);
         }
-        tables[i].columns = strdup(sql);
+
+        int i_hstore_column;
+        for(i_hstore_column = 0; i_hstore_column < Options->n_hstore_columns; i_hstore_column++)
+        {
+            strcat(sql, ",\"");
+            strcat(sql, Options->hstore_columns[i_hstore_column]);
+            strcat(sql, "\" ");
+        }
+    
+	if (Options->enable_hstore) strcat(sql,",tags");
+
+	tables[i].columns = strdup(sql);
         pgsql_exec(sql_conn, PGRES_COPY_IN, "COPY %s (%s,way) FROM STDIN", tables[i].name, tables[i].columns);
 
         tables[i].copyMode = 1;
@@ -1149,10 +1342,10 @@ static void *pgsql_out_stop_one(void *arg)
       pgsql_exec(sql_conn, PGRES_COMMAND_OK, "CREATE TABLE %s_tmp AS SELECT * FROM %s ORDER BY way;\n", table->name, table->name);
       pgsql_exec(sql_conn, PGRES_COMMAND_OK, "DROP TABLE %s;\n", table->name);
       pgsql_exec(sql_conn, PGRES_COMMAND_OK, "ALTER TABLE %s_tmp RENAME TO %s;\n", table->name, table->name);
-      pgsql_exec(sql_conn, PGRES_COMMAND_OK, "CREATE INDEX %s_index ON %s USING GIST (way GIST_GEOMETRY_OPS);\n", table->name, table->name);
+      pgsql_exec(sql_conn, PGRES_COMMAND_OK, "CREATE INDEX %s_index ON %s USING GIST (way GIST_GEOMETRY_OPS) TABLESPACE %s;\n", table->name, table->name, Options->tblsindex);
       /* slim mode needs this to be able to apply diffs */
       if( Options->slim )
-          pgsql_exec(sql_conn, PGRES_COMMAND_OK, "CREATE INDEX %s_pkey ON %s USING BTREE (osm_id);\n", table->name, table->name);
+	pgsql_exec(sql_conn, PGRES_COMMAND_OK, "CREATE INDEX %s_pkey ON %s USING BTREE (osm_id) TABLESPACE %s;\n", table->name, table->name, Options->tblsindex);
       pgsql_exec(sql_conn, PGRES_COMMAND_OK, "GRANT SELECT ON %s TO PUBLIC;\n", table->name);
       pgsql_exec(sql_conn, PGRES_COMMAND_OK, "ANALYZE %s;\n", table->name);
     }
@@ -1424,18 +1617,18 @@ static int pgsql_modify_relation(int osm_id, struct member *members, int member_
 }
 
 struct output_t out_pgsql = {
-        start:     pgsql_out_start,
-        stop:      pgsql_out_stop,
-        cleanup:   pgsql_out_cleanup,
-        node_add:      pgsql_add_node,
-        way_add:       pgsql_add_way,
-        relation_add:  pgsql_add_relation,
+        .start           = pgsql_out_start,
+        .stop            = pgsql_out_stop,
+        .cleanup         = pgsql_out_cleanup,
+        .node_add        = pgsql_add_node,
+        .way_add         = pgsql_add_way,
+        .relation_add    = pgsql_add_relation,
         
-        node_modify: pgsql_modify_node,
-        way_modify: pgsql_modify_way,
-        relation_modify: pgsql_modify_relation,
+        .node_modify     = pgsql_modify_node,
+        .way_modify      = pgsql_modify_way,
+        .relation_modify = pgsql_modify_relation,
 
-        node_delete: pgsql_delete_node,
-        way_delete: pgsql_delete_way,
-        relation_delete: pgsql_delete_relation
+        .node_delete     = pgsql_delete_node,
+        .way_delete      = pgsql_delete_way,
+        .relation_delete = pgsql_delete_relation
 };
