@@ -22,11 +22,16 @@ struct output_options {
   int slim;        /* In slim mode */
   int cache;       /* Memory usable for cache in MB */
   struct middle_t *mid;  /* Mid storage to use */
+  const char *tblsindex;     /* Pg Tablespace to store indexes */
   const char *style;     /* style file to use */
   int expire_tiles_zoom;	/* Zoom level for tile expiry list */
   int expire_tiles_zoom_min;	/* Minimum zoom level for tile expiry list */
   const char *expire_tiles_filename;	/* File name to output expired tiles list to */
+  int enable_hstore; /* add an additional hstore column with objects key/value pairs */
   int enable_multi; /* Output multi-geometries intead of several simple geometries */
+  const char** hstore_columns; /* list of columns that should be written into their own hstore column */
+  int n_hstore_columns; /* number of hstore columns */
+  int keep_coastlines;
 };
 
 struct output_t {
