@@ -9,6 +9,7 @@
 #include <string.h>
 #include <unistd.h>
 #include <proj_api.h>
+#include <math.h>
 
 #include "reprojection.h"
 
@@ -32,20 +33,20 @@ static int Proj;
 
 const struct Projection_Info Projection_Info[] = {
   [PROJ_LATLONG] = { 
-     descr: "Latlong", 
-     proj4text: "+init=epsg:4326",
-     srs:4326, 
-     option: "-l" },
-  [PROJ_MERC]    = { 
-     descr: "WGS84 Mercator", 
-     proj4text: "+proj=merc +datum=WGS84  +k=1.0 +units=m +over +no_defs", 
-     srs:3395, 
-     option: "-M" },
+     .descr     = "Latlong", 
+     .proj4text = "+init=epsg:4326",
+     .srs       = 4326, 
+     .option    = "-l" },
+  [PROJ_MERC] = { 
+     .descr     = "WGS84 Mercator", 
+     .proj4text = "+proj=merc +datum=WGS84  +k=1.0 +units=m +over +no_defs", 
+     .srs       = 3395, 
+     .option    = "-M" },
   [PROJ_SPHERE_MERC] = { 
-     descr: "Spherical Mercator",  
-     proj4text: "+proj=merc +a=6378137 +b=6378137 +lat_ts=0.0 +lon_0=0.0 +x_0=0.0 +y_0=0 +k=1.0 +units=m +nadgrids=@null +wktext  +no_defs", 
-     srs:900913, 
-     option: "-m" }
+     .descr     = "Spherical Mercator",  
+     .proj4text = "+proj=merc +a=6378137 +b=6378137 +lat_ts=0.0 +lon_0=0.0 +x_0=0.0 +y_0=0 +k=1.0 +units=m +nadgrids=@null +wktext  +no_defs", 
+     .srs       = 900913, 
+     .option    = "-m" }
 };
 static struct Projection_Info custom_projection;
 
@@ -137,6 +138,22 @@ void reproject(double *lat, double *lon)
     if (Proj == PROJ_LATLONG)
         return;
 
+    if (Proj == PROJ_SPHERE_MERC)
+    {
+      /* The latitude co-ordinate is clipped at slightly larger than the 900913 'world' 
+       * extent of +-85.0511 degrees to ensure that the points appear just outside the
+       * edge of the map. */
+
+        if (*lat > 85.07)
+            *lat = 85.07;
+        if (*lat < -85.07)
+            *lat = -85.07;
+
+        *lat = log(tan(M_PI/4.0 + (*lat) * DEG_TO_RAD / 2.0)) * EARTH_CIRCUMFERENCE/(M_PI*2);
+        *lon = (*lon) * EARTH_CIRCUMFERENCE / 360.0;
+        return;
+    }
+
     x[0] = *lon * DEG_TO_RAD;
     y[0] = *lat * DEG_TO_RAD;
     z[0] = 0;
@@ -145,7 +162,7 @@ void reproject(double *lat, double *lon)
     
     pj_transform(pj_source, pj_target, 1, 1, x, y, z);
     
-    //printf("%.4f\t%.4f -> %.4f\t%.4f\n", *lat, *lon, y[0], x[0]);    
+    //printf("%.4f\t%.4f -> %.4f\t%.4f\n", *lat, *lon, y[0], x[0]);
     *lat = y[0];
     *lon = x[0];
 }
